/*
** Module   : handle.java
** Abstract : Progress 4GL compatible handle object
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 GES 20051125   @23505 Created initial version which represents
**                           the Progress handle data type.
** 002 GES 20060206   @24300 Made compareTo() safe to process any object
**                           type (avoids ClassCastException).
** 003 GES 20060208   @24390 Added a default format string getter.
** 004 GES 20060724   @28161 Added a forced assign() abstract method.
** 005 GES 20060807   @28479 Moved to externalizable interface to optimize
**                           network performance. Throw an exception if
**                           an attempt is made to serialize this class
**                           since it makes no sense to transfer this
**                           between systems.
** 006 GES 20070219   @32175 Added support for the widget-handle built-in
**                           function.
** 007 GES 20070221   @32195 Added support for the TransparentWrapper as
**                           a contained object that exposes its
**                           referenced object instead.
** 008 GES 20080402   @37825 Added a new assign(Object) signature and a
**                           new constructor handle(WrappedResource).
** 009 CA  20090227   @41383 Undefined SELF or FOCUS handles refers a
**                           NullWidget instance - modified the isUnknown
**                           method to return true when the referenced
**                           value is not null, but unknown.
** 010 ECF 20101003          Removed ContextContainer inner class; replaced with
**                           parameterized use of ContextLocal. Fixed javadoc.
** 011 OM  20130121          Added stub for DELETE OBJECT implementation.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;
import com.goldencode.p2j.security.ContextLocal;

/**
 * A class that represents a Progress 4GL compatible handle data type. This
 * is a container for an object of type {@link WrappedResource}.  Such
 * objects can be passed around as a regular {@link BaseDataType} wrapper
 * but the backing object can be indirectly accessed via the {@link #get} 
 * method.
 * <p>
 * The following is the mapping of Progress language features to the
 * corresponding feature in this class:
 * <p>
 * <pre>
 * string        function (default format)      {@link #toString()}
 *                        (user-defined format) {@link #toString(String)} 
 *                        (export format)       {@link #toStringExport()}
 * valid-handle  function                       {@link #isValid}
 * widget-handle function                       {@link #fromString}
 * </pre>
 * <p>
 * There are 2 deviations between this class and its behavior in the Progress
 * 4GL:
 * <p>
 * <ol>
 *    <li> The default format string is larger in Java than in Progress since
 *         the Java implementation uses all possible 32-bit values as its
 *         handle identifiers. This could be made compatible by limiting the
 *         possible values used to those that would fit into the Progress
 *         equivalent. This would probably require that every handle be
 *         registered or assigned a unique value at construction. 
 *    <li> In Progress, it is possible that one could create a loop to 
 *         instantiate all possible handles using the WIDGET-HANDLE function.
 *         This means that one might not have used the STRING function to get
 *         the text representation since that loop could create strings for
 *         all integers in the possible range of handle identifiers. This
 *         seems like a bad practice, but if it is ever used in reality, it
 *         would invalidate the implementation of {@link #fromString} which
 *         depends upon a map maintained during {@link #toString}.  The
 *         solution is the same as for the oversized text representation
 *         problem above (a registry created at construction of every 
 *         handle).
 * </ol>
 */
public class handle
extends BaseDataType
{
   /** The default format string (WARNING: not compatible with Progress). */
   private static final String defaultFormat = ">>>>>>>>>9";   
   
   /** Stores context-local state variables. */
   private static final ContextLocal<WorkArea> work =
   new ContextLocal<WorkArea>()
   {
      protected synchronized WorkArea initialValue()
      {
         return new WorkArea();
      }
   };
      
   /** Stores the referenced data of this handle. */
   private WrappedResource value = null;
   
   /**
    * Default constructor, creates an instance that represents the unknown
    * value.
    */
   public handle()
   {
   }
   
   /**
    * Constructs an instance that has the contained object reference and
    * <code>unknown value</code> that exactly matches that of the passed-in 
    * instance.
    *
    * @param    value
    *           The instance to pattern from.
    */
   public handle(handle value)
   {
      assign(value, false);
   }
   
   /**
    * Constructs an instance that has the contained object reference.
    *
    * @param    value
    *           The instance to pattern from.
    */
   public handle(WrappedResource value)
   {
      assign(value, false);
   }
   
   /**
    * Convert the string representation of a handle into the original handle
    * instance. This facility can only work for handles that have been
    * converted at least once into a string via {@link #toString}. Such
    * handles are stored in a weak hash map for conversion purposes.
    *
    * @param    txt
    *           A valid text representation of a handle.
    */
   public static handle fromString(character txt)
   {
      return fromString(txt.toStringMessage());
   }
   
   /**
    * Convert the string representation of a handle into the original handle
    * instance. This facility can only work for handles that have been
    * converted at least once into a string via {@link #toString}. Such
    * handles are stored in a weak hash map for conversion purposes.
    *
    * @param    txt
    *           A valid text representation of a handle.
    */
   public static handle fromString(String txt)
   {
      handle result = null;
      
      try
      {
         Integer key = new Integer(txt);
         
         WorkArea wa = work.get();
         result = (handle) wa.map.get(key);
      }
      
      catch (NumberFormatException nfe)
      {
         // return an invalid handle
         return new handle();
      }
      
      return result;
   }
   
   /**
    * Creates a new instance of the same type that represents the 
    * <code>unknown value</code>.
    *
    * @return   An instance that represents the <code>unknown value</code>.
    */
   public BaseDataType instantiateUnknown()   
   {
      return new handle();
   }
   
   /**
    * Does the same as standard <code>clone()</code> method but returns an
    * instance of <code>BaseDataType</code> and doesn't throw the
    * <code>CloneNotSupportedException</code>.
    *
    * @return   A clone of this instance.  
    */
   public BaseDataType duplicate()
   {
      return new handle(this);
   }
   
   /**
    * Hash code implementation which is consistent with {@link
    * BaseDataType#equals}.
    *
    * @return  Hash code value for this object instance.
    */
   public int hashCode()
   {  
      return value == null ? ((Object) this).hashCode() : value.hashCode();
   }
   
   /**
    * Returns the value of this instance as an <code>Object</code>.
    * <p>
    * If this instance represents the <code>unknown value</code>, a
    * <code>null</code> will be returned.
    *
    * @return   The value as an <code>Object</code> or <code>null</code> if
    *           this is <code>unknown</code>.
    */
   public Object get()
   {
      Object result = value;
      
      // redirect to contained object if it is the special wrapper
      if (result != null && result instanceof TransparentWrapper)
      {
         TransparentWrapper wrap = (TransparentWrapper) result;
         result = wrap.get();
      }
      
      return result;
   }   
   
   /**
    * Set the value of this instance as a {@link WrappedResource}.
    *
    * @param    value
    *           The value to be represented in this instance.
    */
   public void set(WrappedResource value)
   {
      this.value = value;
   }
   
   /**
    * Determines if this instance is valid (to be used).
    *
    * @return   <code>true</code> if this instance can be used.
    */
   public logical isValid()
   {
      boolean result = (value == null) ? false : value.isValid();
      
      return new logical(result);
   }   
   
   /**
    * Determines if this instance represents the <code>unknown value</code>.
    *
    * @return   <code>true</code> if this instance is set to the
    *           <code>unknown value</code>.
    */
   public boolean isUnknown()
   {
      return (value == null || value.isUnknown());
   }
   
   /**
    * Sets the state of this instance's <code>unknown value</code> flag to
    * <code>true</code>.
    * <p>
    * <b>Warning: a separate call is needed to ensure that the data of this
    * instance is set to the correct value.</b>
    */
   public void setUnknown()
   {
      value = null;
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    * @param    force
    *           Flag to indicate if a pending error should be ignored (see 
    *           {@link ErrorManager#isPendingError} and the assignment
    *           forced regardless.  Note that normal assignment silently
    *           exits with no change if there is a pending error. If this
    *           flag is <code>true</code>, that check is bypassed and the
    *           assignment will always be done.
    */
   public void assign(BaseDataType value, boolean force)
   {
      assign((handle) value, force);
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    * @param    force
    *           Flag to indicate if a pending error should be ignored (see 
    *           {@link ErrorManager#isPendingError} and the assignment
    *           forced regardless.  Note that normal assignment silently
    *           exits with no change if there is a pending error. If this
    *           flag is <code>true</code>, that check is bypassed and the
    *           assignment will always be done.
    */
   public void assign(handle value, boolean force)
   {
      assign(value.value, force);
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    * @param    force
    *           Flag to indicate if a pending error should be ignored (see 
    *           {@link ErrorManager#isPendingError} and the assignment
    *           forced regardless.  Note that normal assignment silently
    *           exits with no change if there is a pending error. If this
    *           flag is <code>true</code>, that check is bypassed and the
    *           assignment will always be done.
    */
   public void assign(WrappedResource value, boolean force)
   {
      if (force || !ErrorManager.isPendingError())
      {
         this.value = value;
      }
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    * <p>
    * This is a "no operation" in the case where silent error mode is active
    * AND an error has occurred in the current processing since the opening
    * call to {@link ErrorManager#silentErrorEnable}. This protection is
    * needed to ensure that the state of this instance does not change in
    * this case.  This will not affect undo because that only occurs when
    * silent error mode is disabled.
    *
    * @param    value
    *           The instance from which to copy state. This must be of
    *           type <code>handle</code> or <code>WrappedResource</code>.
    */
   public void assign(Object value)
   {
      if (value instanceof WrappedResource)
      {
         assign((WrappedResource) value);
      }
      else if (value instanceof handle)
      {
         assign((handle) value);
      }
      else
      {
         // unexpected failure
         throw new IllegalArgumentException("Invalid value type.");
      }
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    * <p>
    * This is a "no operation" in the case where silent error mode is active
    * AND an error has occurred in the current processing since the opening
    * call to {@link ErrorManager#silentErrorEnable}. This protection is
    * needed to ensure that the state of this instance does not change in
    * this case.  This will not affect undo because that only occurs when
    * silent error mode is disabled.
    *
    * @param    value
    *           The instance from which to copy state.
    */
   public void assign(handle value)
   {
      assign(value.value, false);
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    */
   public void assign(Undoable value)
   {
      assign((handle) value, false);
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * given value.  Note that <code>null</code> may be passed in which case
    * this instance will be set to the <code>unknown value</code>.
    * <p>
    * This is a "no operation" in the case where silent error mode is active
    * AND an error has occurred in the current processing since the opening
    * call to {@link ErrorManager#silentErrorEnable}. This protection is
    * needed to ensure that the state of this instance does not change in
    * this case.  This will not affect undo because that only occurs when
    * silent error mode is disabled.
    *
    * @param    value
    *           The instance from which to copy state.
    */
   public void assign(WrappedResource value)
   {
      assign(value, false);
   }
   
   /**
    * Compares this instance with the specified instance and returns a -1
    * if this instance is less than the specified, 0 if the two instances
    * are equal and 1 if this instance is greater than the specified
    * instance.  This is the implementation of the <code>Comparable</code>
    * interface.
    * <p>
    * The algorithm will fail to give meaningful results in the case where
    * one tries to sort against other objects that do not represent compatible
    * values.
    *
    * @param    obj
    *           The instance to compare against.
    *
    * @return   0 or 1 depending on if the passed instance is or is not
    *           <code>unknown</code>.
    */
   public int compareTo(Object obj)   
   {  
      handle h = null;
      
      if (obj instanceof handle)
      {
         h = (handle) obj;
      }
      else
      {
         return -1;
      }
      
      int result = 0;
      
      int h1 = hashCode();
      int h2 = h.hashCode();      
      
      if (h1 < h2)
      {
         result = -1;
      }
      else if (h1 == h2)
      {
         if (value.equals(h.value))
         {
            result = 0;
         }
         
         // arbitrary but if the hashcodes are equal and the instances are
         // not actually equal, then make sure this is clear
         result = 1;
      }
      else
      {
         result = 1;
      }
         
      return result;
   }
   
   /**
    * Creates a string representation of the instance data using the default
    * Progress 4GL format of '>>>>>>>>>9'.  If the instance represents the
    * <code>unknown value</code>, a '?' will be returned.
    *
    * @return   The formatted string.
    */
   public String toString()   
   {
      return toString(null);
   }
   
   /**
    * Creates a string representation of the instance data using the user
    * specified format string (Progress 4GL compatible) or if no such
    * string is provided, the default format of '>>>>>>>>>9'.  If the
    * instance represents the <code>unknown value</code>, a '?' will be
    * returned.
    *
    * @param    fmt
    *           The Progress 4GL format string or <code>null</code> if the
    *           default format is to be used.
    *
    * @return   The formatted string.
    */
   public String toString(String fmt)   
   {
      if (isUnknown())
      {
         return "?";
      }
      
      int hc = hashCode();
      
      // this handle's text value is being requested, save off the reference
      // in a weak hash map so that later, it can be converted back via the
      // widgetHandle() method
      WorkArea wa = work.get();
      wa.map.put(new Integer(hc), this);
      
      integer i = new integer(hc);
      return i.toString((fmt == null) ? defaultFormat : fmt);
   }
   
   /**
    * Creates a string representation of the instance data in a form that is
    * compatible with the <code>MESSAGE</code> language statement.  If the
    * instance represents the <code>unknown value</code>, a '?' will be
    * returned.
    *
    * @return   The 'message' formatted string.
    */
   public String toStringMessage()   
   {
      return toString();
   }
   
   /**
    * Creates a string representation of the instance data using the 'export'
    * format.  If the instance represents the <code>unknown value</code>, a
    * '?' will be returned.
    *
    * @return   The 'export' formatted string.
    */
   public String toStringExport()   
   {
      return toString();
   }   
   
   /**
    * Return the default display format string for this type.
    * <p>
    * Warning: this is larger than the Progress version because the range of
    * handle values can be a 32-bit integer in this implementation.
    *
    * @return   The default format string.
    */
   public String defaultFormatString()   
   {
      return defaultFormat;
   }      
      
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      throw new NotSerializableException("handle");
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      throw new NotSerializableException("handle");
   }

   /**
    * This is the equivalent of <code>DELETE OBJECT<code> statement stub.
    * Deletes an object such as a widget, a procedure, a server, a socket, or 
    * an instance of a class. Deleting the object causes all allocated 
    * resources associated with the object to be returned to the system 
    * (except when otherwise noted).
    */
   public void delete()
   {
      // TODO Implement this method 
   }

   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** 
       * Stores handles that have been converted to strings and may need
       * to be converted back into a handle.
       */
      private Map<Integer, handle> map = new HashMap<Integer, handle>();
   }
}
